/*
** clock.c - Example showing how to program the clock generators on an
**           ADM-XRC series card.
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRCII-Lite
**   o ADM-XRCII
**   o ADM-XRCIIPro-Lite
**   o ADP-WRCII
**   o ADP-DRCII
**
** (C) Copyright 2000-2004 Alpha Data Parallel Systems Ltd.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>

#include <admxrc2.h>

#include "../common/common.h"

int main(int argc, char *argv[])
{
  int               ret = 0;
  ADMXRC2_STATUS    status;
  ADMXRC2_HANDLE    card = ADMXRC2_HANDLE_INVALID_VALUE;
  ADMXRC2_CARD_INFO cardInfo;
  double            freq;
  double            afreq;
  unsigned long     index;
  Option            options[2];
  Arguments         arguments;
  OptionType        paramType[2];
  char*             paramInfo[2];
  char*             paramMeaning[2];

  options[0].key = "card";
  options[0].type = Option_uint;
  options[0].def.uintVal = 0;
  options[0].meaning = "ID of card to use";
  options[1].key = "index";
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = "index of card to use";
  arguments.nOption = sizeof(options) / sizeof(Option);
  arguments.option = options;

  paramType[0] = Option_int;
  paramInfo[0] = "clk_idx";
  paramMeaning[0] = "Clock index";
  paramType[1] = Option_double;
  paramInfo[1] = "f_clk";
  paramMeaning[1] = "Desired clock frequency (MHz)";
  arguments.nParamType = 2;
  arguments.minNParam = 2;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;

  parse_command_line("clock", argc, argv, &arguments);

  if (options[0].specified && options[1].specified) {
    printf("Do not specify both /card and /index\n");
    ret = -1;
    goto done;
  }

  if (!options[1].specified) {
    ADMXRC2_CARDID cardID = options[0].value.uintVal;

    status = ADMXRC2_OpenCard(cardID, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to open card with ID %ld: %s\n",
	     (unsigned long) cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[1].value.uintVal;

    status = ADMXRC2_OpenCardByIndex(index, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to open card with index %ld: %s\n",
	     (unsigned long) index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  index = arguments.param[0].uintVal;
  freq = arguments.param[1].doubleVal * 1000000.0;

  status = ADMXRC2_GetCardInfo(card, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to get card info: %s\n", ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADMXRC2_PRO_LITE:
  case ADMXRC2_BOARD_ADPWRC2:
  case ADMXRC2_BOARD_ADPDRC2:
    break;

  default:
    printf("This example must be run on one of the following cards:\n\n");
    printf("\tADM-XRC\n");
    printf("\tADM-XRC-P\n");
    printf("\tADM-XRCII-Lite\n");
    printf("\tADM-XRCII\n");
    printf("\tADM-XRCIIPro-Lite\n");
    printf("\tADP-WRCII\n");
    printf("\tADP-DRCII\n");
    ret = -1;
    goto done;
  }

  /*
  ** Check that the clock generator index specified is within range for
  ** the card being used.
  */
  if (index >= cardInfo.NumClock) {
    printf("Invalid clock index: %ld\n", index);
    ret = -1;
    goto done;
  }

  /*
  ** Set the frequency of the specified clock generator. The
  ** actual programmed frequency, which may be slightly different
  ** from the desired frequency, is returned is 'afreq'.
  */
  status = ADMXRC2_SetClockRate(card, index, freq, &afreq);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to set clock %ld to %.1fMHz: %s\n",
	   index, freq / 1.0e6, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  printf("Actual frequency of clock %ld = %fHz\n", index, afreq);

 done:
  if (card != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(card);
  }

  return ret;
}
