/*
** memtest.c - Example demonstrating host access to the SSRAM
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRCII-Lite
**   o ADM-XRCII
**   o ADM-XRCIIPro-Lite
**
** (C) Copyright 2001-2003 Alpha Data Parallel Systems Ltd.
**
** This application performs a simple memory test on the SSRAM on
** an ADM-XRC series card, using both programmed I/O and high
** performance DMA.
**
** The 'ZBT' sample FPGA design is loaded, which maps the SSRAM
** into the upper 2MB of FPGA local bus space and some registers
** into the lower 2MB of FPGA local bus space. A page register is
** used in order to give the host to access all of the SSRAM
** (typically 6MB in an ADM-XRCII card).
**
** This application and the 'ZBT' sample FPGA support pipelined or
** flowthrough SSRAMs with 17 to 20 address bits and a width of
** 16(18) or 32(36) bits.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>

#include <admxrc2.h>

#include "../common/common.h"

#define ZBT_PAGE_REG		(0)
#define ZBT_PIPELINE_REG	(1)
#define ZBT_SIZE_REG		(2)
#define ZBT_INFO_REG		(4)
#define ZBT_STATUS_REG		(5)

const unsigned long       ssramOffset = 0x200000U;	 /* In bytes     */
const unsigned long       pageSize    = 0x200000U / 4; /* In longwords */

static unsigned long      numBank;
static unsigned long      bankSize;
static unsigned long      memorySize;
static unsigned long      lastPageSize;

static uint32_t           bankTestMask;
static unsigned long      testSize;

static volatile uint32_t* fpgaSpace;
static volatile uint32_t* ssram;

static uint32_t*          rambuf = NULL;
static uint32_t*          chkbuf = NULL;

static ADMXRC2_HANDLE     card = ADMXRC2_HANDLE_INVALID_VALUE;
static ADMXRC2_DMADESC    dmaDesc1;
static BOOLEAN            dmaDesc1Valid = FALSE;
static ADMXRC2_DMADESC    dmaDesc2;
static BOOLEAN            dmaDesc2Valid = FALSE;
static uint32_t           dmaMode;

static ADMXRC2_STATUS
reverse(uint32_t v)
{
  uint32_t t = 0;
  int      i;
	
  for (i = 0; i < 32; i++) {
    t <<= 1;
    t |= v & 0x1U;
    v >>= 1;
  }
	
  return t;
}

static ADMXRC2_STATUS
readSSRAM(
  const void*   buffer,
  unsigned long	offset,
  unsigned long	length,
  int           useDMA)
{
  ADMXRC2_STATUS     status;
  uint32_t*          d = (uint32_t*) buffer;
  volatile uint32_t* s;
  uint32_t           pgidx;
  uint32_t           pgoffs;
  uint32_t           i, chunk;
	
  while (length) {
    pgidx = (uint32_t) offset & ~(pageSize - 1);
    pgoffs = (uint32_t) offset & (pageSize - 1);
		
    chunk = pageSize - pgoffs;
    if ((pgidx + chunk) > memorySize) {
      chunk = lastPageSize - pgoffs;
    }
    if (chunk > length) {
      chunk = length;
    }

    /* Set the page register */
    fpgaSpace[ZBT_PAGE_REG] = pgidx / pageSize;
    fpgaSpace[ZBT_PAGE_REG]; /* make sure it's got there */
		
    if (useDMA) {
      status = ADMXRC2_DoDMA(card,
			     dmaDesc2,
			     offset * sizeof(uint32_t),
			     chunk * sizeof(uint32_t),
			     ssramOffset + (pgoffs * sizeof(uint32_t)),
			     ADMXRC2_LOCALTOPCI,
			     ADMXRC2_DMACHAN_ANY,
			     dmaMode,
			     0,
			     NULL,
			     NULL);
      if (status != ADMXRC2_SUCCESS) {
	return status;
      }
    } else {
      s = ssram + pgoffs;
      for (i = 0; i < chunk; i++) {
	*d++ = *s++;
      }
    }
		
    offset += chunk;
    length -= chunk;
  }
	
  return ADMXRC2_SUCCESS;
}

static ADMXRC2_STATUS
writeSSRAM(
  const void*   buffer,
  unsigned long offset,
  unsigned long length,
  int           useDMA)
{
  ADMXRC2_STATUS     status;
  volatile uint32_t* d;
  uint32_t*          s = (uint32_t*) buffer;
  uint32_t           pgidx;
  uint32_t           pgoffs;
  uint32_t           i, chunk;
	
  while (length) {
    pgidx = (uint32_t) offset & ~(pageSize - 1);
    pgoffs = (uint32_t) offset & (pageSize - 1);
		
    chunk = pageSize - pgoffs;
    if ((pgidx + chunk) > memorySize) {
      chunk = lastPageSize - pgoffs;
    }
    if (chunk > length) {
      chunk = length;
    }
			
    /* Set the page register */
    fpgaSpace[ZBT_PAGE_REG] = pgidx / pageSize;
    fpgaSpace[ZBT_PAGE_REG]; /* make sure it's got there */
		
    if (useDMA) {
      status = ADMXRC2_DoDMA(card,
			     dmaDesc1,
			     offset * sizeof(uint32_t),
			     chunk * sizeof(uint32_t),
			     ssramOffset + (pgoffs * sizeof(uint32_t)),
			     ADMXRC2_PCITOLOCAL,
			     ADMXRC2_DMACHAN_ANY,
			     dmaMode,
			     0,
			     NULL,
			     NULL);
      if (status != ADMXRC2_SUCCESS) {
	return status;
      }
    } else {
      d = ssram + pgoffs;
      for (i = 0; i < chunk; i++) {
	*d++ = *s++;
      }
      *(d - 1); /* make sure it's got there */
    }
		
    offset += chunk;
    length -= chunk;
  }
	
  return ADMXRC2_SUCCESS;
}

int main(int argc, char *argv[])
{
  int                ret = 0;
  const float        mclk_freq = 33.0f;
  ADMXRC2_STATUS     status;
  ADMXRC2_CARD_INFO  cardInfo;
  ADMXRC2_SPACE_INFO spInfo;
  ADMXRC2_BANK_INFO  bankInfo, bankInfo2;
  const char*        filename;
  Option             options[8];
  Arguments          arguments;
  float              freq, min_freq, max_freq;
  unsigned long      i;
  uint32_t           sizeReg;
  BOOLEAN            dma;
  BOOLEAN            use64bit;
  BOOLEAN            speed;
  unsigned long      nrep;
  BOOLEAN            failed;

  /* Set up expected options */
  options[0].key = "banks";
  options[0].type = Option_hex;
  options[0].def.hexVal = 0xffffffff;
  options[0].meaning = "bitmask of banks to test";
  options[1].key = "card";
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = "ID of card to use";
  options[2].key = "index";
  options[2].type = Option_uint;
  options[2].def.uintVal = 0;
  options[2].meaning = "index of card to use";
  options[3].key = "lclk";
  options[3].type = Option_float;
  options[3].def.floatVal = 33.0;
  options[3].meaning = "local bus clock speed (MHz)";
  options[4].key = "repeat";
  options[4].type = Option_uint;
  options[4].def.uintVal = 1;
  options[4].meaning = "number of repetitions";
  options[5].key = "speed";
  options[5].type = Option_boolean;
  options[5].def.booleanVal = TRUE;
  options[5].meaning = "TRUE => measure access speed";
  options[6].key = "usedma";
  options[6].type = Option_boolean;
  options[6].def.booleanVal = TRUE;
  options[6].meaning = "TRUE => use 64 bit local bus";
  options[7].key = "64";
  options[7].type = Option_boolean;
  options[7].def.booleanVal = FALSE;
  options[7].meaning = "TRUE => use DMA for test";
  arguments.nOption = 8;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;
	
  /* Parse command line */
  parse_command_line("memtest", argc, argv, &arguments);
	
  if (options[1].specified && options[2].specified) {
    printf("*** Do not specify both /card and /index\n");
    ret = -1;
    goto done;
  }
	
  if (!options[2].specified) {
    ADMXRC2_CARDID cardID = options[1].value.uintVal;
		
    status = ADMXRC2_OpenCard(cardID, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Failed to open card with ID %ld: %s\n",
	     (unsigned long) cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[2].value.uintVal;
		
    status = ADMXRC2_OpenCardByIndex(index, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Failed to open card with index %ld: %s\n",
	     (unsigned long) index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }
	
  freq = options[3].value.floatVal;
  nrep = options[4].value.uintVal;
  speed = options[5].value.booleanVal;
  dma = options[6].value.booleanVal;
  use64bit = options[7].value.booleanVal;

  /* 
  ** Get card information, specifically for device fitted and
  ** number of SSRAM banks
  */
  status = ADMXRC2_GetCardInfo(card, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to get card info: %s\n", ADMXRC2_GetStatusString(status));
    return -1;
  }
	
  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
    min_freq = 25.0;
    max_freq = 40.0;
    break;

  case ADMXRC2_BOARD_ADMXRC2:
    min_freq = 24.0;
    max_freq = 66.0;
    break;

  case ADMXRC2_BOARD_ADMXRC2_PRO_LITE:
    min_freq = 24.0;
    max_freq = 80.0;
    break;

  default:
    printf("*** This example must be run on one of the following cards:\n\n");
    printf("\tADM-XRC\n");
    printf("\tADM-XRC-P\n");
    printf("\tADM-XRCII-Lite\n");
    printf("\tADM-XRCII\n");
    printf("\tADM-XRCIIPro-Lite\n");
    ret = -1;
    goto done;
  }
	
  if (freq > max_freq || freq < min_freq) {
    printf("*** LCLK frequency of %.1f MHz is not supported\n", freq);
    ret = -1;
    goto done;
  }
	
  if (use64bit && cardInfo.BoardType != ADMXRC2_BOARD_ADMXRC2_PRO_LITE) {
    printf("*** 64-bit local bus is not supported on the specified card.\n");
    ret = -1;
    goto done;
  }

  /* Get the address of FPGA space */
  status = ADMXRC2_GetSpaceInfo(card, 0, &spInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to get space 0 info: %s\n",
	   ADMXRC2_GetStatusString(status));
    return -1;
  }
  fpgaSpace = (volatile uint32_t*) spInfo.VirtualBase;
	
  if (!options[0].specified) {
    bankTestMask = cardInfo.RAMBanksFitted;
  } else {
    bankTestMask = options[0].value.hexVal & cardInfo.RAMBanksFitted;
  }

  /*
  ** Check that the type, size, width etc. of each SSRAM bank
  ** is the same.
  */
  numBank = 0;
  testSize = 0;
  for (i = 0; i < cardInfo.NumRAMBank; i++) {
    status = ADMXRC2_GetBankInfo(card, i, &bankInfo2);
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Failed to get bank %ld info: %s\n",
	     (unsigned long) i,
	     ADMXRC2_GetStatusString(status));
      return -1;
    }

    if (!(bankInfo2.Type & (ADMXRC2_RAM_ZBTP | ADMXRC2_RAM_ZBTFT))) {
      /* Ignore any banks that are not ZBT SSRAM */
      continue;
    }

    if (!bankInfo2.Fitted) {
      printf("*** Not all SSRAM banks are fitted - aborting.\n");
      ret = -1;
      goto done;
    }

    if (cardInfo.BoardType == ADMXRC2_BOARD_ADMXRC2_PRO_LITE && !use64bit) {
      /*
      ** ADM-XPL ZBT bank width is 64 bits but the 32-bit version of design
      ** uses only the lower 32 bits of the ZBT bank.
      */
      bankInfo2.Width = 32;
    }

    if (!numBank) {
      memcpy(&bankInfo, &bankInfo2, sizeof(bankInfo));
      bankSize = (bankInfo.Width / 8) * bankInfo.Size / 4;
    } else {
      if (bankInfo2.Type != bankInfo.Type ||
	  bankInfo2.Width != bankInfo.Width ||
	  bankInfo2.Size != bankInfo.Size) {
	printf("*** Not all SSRAM banks are the same - aborting.\n");
	ret = -1;
	goto done;
      }
    }

    numBank++;

    if (bankTestMask & (0x1UL << i)) {
      testSize += (bankInfo.Width / 8) * bankInfo.Size / 4;
    }
  }
	
  if (!numBank) {
    printf("*** No SSRAM banks found - aborting.\n");
    ret = -1;
    goto done;
  }

  /*
  ** Calculate memory size in longwords (32-bit words)
  */
  memorySize = numBank * bankSize;
  lastPageSize = memorySize & (pageSize - 1);

  /*
  ** Determine value to program into the SIZE register in the FPGA
  */
  switch (bankInfo.Size) {
  case 128*1024:
    sizeReg = 0;
    break;
  case 256*1024:
    sizeReg = 1;
    break;
  case 512*1024:
    sizeReg = 2;
    break;
  case 1024*1024:
    sizeReg = 3;
    break;
  default:
    printf("*** SSRAM of %ld words x %ld is not supported by this application\n",
	   (unsigned long) bankInfo.Size, (unsigned long) bankInfo.Width);
    ret = -1;
    goto done;
  }
	
  ssram = (volatile uint32_t*) (((uint8_t*) fpgaSpace) + ssramOffset);
	
  rambuf = (uint32_t*) ADMXRC2_Malloc(memorySize * sizeof(uint32_t));
  if (rambuf == NULL) {
    printf("*** Failed to allocate RAM buffer - aborting\n");
    ret = -1;
    goto done;
  }
  chkbuf = (uint32_t*) ADMXRC2_Malloc(memorySize * sizeof(uint32_t));
  if (chkbuf == NULL) {
    printf("*** Failed to allocate RAM readback buffer - aborting\n");
    ret = -1;
    goto done;
  }
	
  status = ADMXRC2_SetClockRate(card, ADMXRC2_CLOCK_LCLK, freq * 1.0e6, NULL);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to set LCLK to %.1fMHz: %s\n",
	   freq, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
	
  if (cardInfo.BoardType != ADMXRC2_BOARD_ADMXRC2_PRO_LITE) {
    status = ADMXRC2_SetClockRate(card, 1, mclk_freq * 1.0e6, NULL);
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Failed to set clock 1 to %.1fMHz: %s\n",
	     mclk_freq, ADMXRC2_GetStatusString(status));
      return -1;
    }
  }

  if (use64bit) {
    filename = get_fpga_filename(&cardInfo, "zbt64", &arguments);
  } else {
    filename = get_fpga_filename(&cardInfo, "zbt", &arguments);
  }
  status = ADMXRC2_ConfigureFromFile(card, filename);
  if (status != ADMXRC2_SUCCESS) {
    printf ("*** Failed to load the bitstream '%s': %s\n",
	    filename, ADMXRC2_GetStatusString(status));
    return -1;
  }
	
  if (use64bit) {
    uint32_t config;
    
    config = ADMXRC2_SPACE_SET_WIDTH | ADMXRC2_SPACE_WIDTH_64;
    status = ADMXRC2_SetSpaceConfig(card, 0, config);
    if (status != ADMXRC2_SUCCESS) {
      printf ("*** Failed to set space configuration: %s\n",
	      ADMXRC2_GetStatusString(status));
      return -1;
    }
  }
  
  status = ADMXRC2_SetupDMA(card,
			    rambuf,
			    memorySize * sizeof(uint32_t),
			    0,
			    &dmaDesc1);
  if (status != ADMXRC2_SUCCESS) {
    printf ("*** Failed to get a DMA descriptor: %s\n",
	    ADMXRC2_GetStatusString(status));
    return -1;
  }
	
  status = ADMXRC2_SetupDMA(card,
			    chkbuf,
			    memorySize * sizeof(uint32_t),
			    0,
			    &dmaDesc2);
  if (status != ADMXRC2_SUCCESS) {
    printf ("*** Failed to get a DMA descriptor: %s\n",
	    ADMXRC2_GetStatusString(status));
    return -1;
  }
	
  dmaMode = ADMXRC2_BuildDMAModeWord(cardInfo.BoardType,
				     use64bit ? ADMXRC2_IOWIDTH_64 : ADMXRC2_IOWIDTH_32,
				     0,
				     ADMXRC2_DMAMODE_USEREADY | ADMXRC2_DMAMODE_USEBTERM | ADMXRC2_DMAMODE_BURSTENABLE);
	
  /* Wait for DLLs/DCMs to lock */
  sleep(1);

  /*
  ** Program the number of address bits used by the SSRAMs
  */
  fpgaSpace[ZBT_SIZE_REG] = sizeReg;
  fpgaSpace[ZBT_SIZE_REG]; /* make sure it's got there */

  printf("Size reg   = 0x%8.8lx\n", (unsigned long) fpgaSpace[ZBT_SIZE_REG]);
  printf("Info reg   = 0x%8.8lx\n", (unsigned long) fpgaSpace[ZBT_INFO_REG]);
  printf("Status reg = 0x%8.8lx\n", (unsigned long) fpgaSpace[ZBT_STATUS_REG]);
	
  failed = FALSE;
	
  printf("Performing memory tests using %s\n",
	 dma ? "DMA" : "programmed I/O");
	
  /* Iterate over flowthrough and pipelined */
  for (i = 0; i < 2; i++) {
    int           nError = 0;
    unsigned long j, k;
    uint32_t      exp, act;
    unsigned long rep;
		
    if (i == 0) {
      if (!(bankInfo.Type & ADMXRC2_RAM_ZBTFT)) {
	continue;
      }
			
      /*
      ** Set the FPGA to flowthrough mode
      */
      fpgaSpace[ZBT_PIPELINE_REG] = 0x0;
      fpgaSpace[ZBT_PIPELINE_REG]; /* make sure it's got there */
      printf("Testing %ld kB in flowthrough mode...\n", testSize * 4 / 1024);
    } else {
      if (!(bankInfo.Type & ADMXRC2_RAM_ZBTP)) {
	continue;
      }
			
      /*
      ** Set the FPGA to pipelined mode
      */
      fpgaSpace[ZBT_PIPELINE_REG] = 0x1;
      fpgaSpace[ZBT_PIPELINE_REG]; /* make sure it's got there */
      printf("Testing %ld kB in pipelined mode...\n", testSize * 4 / 1024);
    }
		
    for (rep = 0; rep < nrep; rep++) {
			
      printf("Repetition %ld\n", rep);
			
      /* Perform 0x55555555 test - look for shorted/stuck data pins */
      if (!failed && nError == 0) {
	printf("Writing 0x55555555...\n");
				
	exp = 0x55555555U;
	for (j = 0; j < memorySize; j++) {
	  rambuf[j] = exp;
	}
				
	for (j = 0; j < numBank; j++) {
	  if (!(bankTestMask & (1U << j))) {
	    continue;
	  }
					
	  status = writeSSRAM(rambuf + j * bankSize, j * bankSize, bankSize, dma);
	  if (status != ADMXRC2_SUCCESS) {
	    printf("Error: failed to write SSRAM\n");
	    failed = TRUE;
	    break;
	  }
	}
				
	if (!failed) {
	  for (j = 0; j < numBank; j++) {
	    if (!(bankTestMask & (1U << j))) {
	      continue;
	    }
						
	    status = readSSRAM(chkbuf + j * bankSize, j * bankSize, bankSize, dma);
	    if (status != ADMXRC2_SUCCESS) {
	      printf("Error: failed to read SSRAM\n");
	      failed = TRUE;
	      break;
	    }
	  }
					
	  for (j = 0; j < numBank; j++) {
	    if (!(bankTestMask & (1U << j))) {
	      continue;
	    }
						
	    for (k = 0; k < bankSize; k++) {
	      unsigned long idx = j * bankSize + k;
							
	      act = chkbuf[idx];
	      if (exp != act) {
		if (nError == 10) {
		  printf("[more errors...]\n");
		} else if (nError < 10) {
		  printf("Error at index 0x%8.8lx: expected/actual = 0x%8.8lx/0x%8.8lx\n",
			 (unsigned long) idx, (unsigned long) exp, (unsigned long) act);
		}
		nError++;
	      }
	    }
	  }
	}
      }
			
      /* Perform 0xAAAAAAA test - look for shorted/stuck data pins */
      if (!failed && nError == 0) {
	printf("Writing 0xAAAAAAAA...\n");
				
	exp = 0xAAAAAAAAU;
	for (j = 0; j < memorySize; j++) {
	  rambuf[j] = exp;
	}
				
	for (j = 0; j < numBank; j++) {
	  if (!(bankTestMask & (1U << j))) {
	    continue;
	  }
					
	  status = writeSSRAM(rambuf + j * bankSize, j * bankSize, bankSize, dma);
	  if (status != ADMXRC2_SUCCESS) {
	    printf("Error: failed to write SSRAM\n");
	    failed = TRUE;
	    break;
	  }
	}
				
	if (!failed) {
	  for (j = 0; j < numBank; j++) {
	    if (!(bankTestMask & (1U << j))) {
	      continue;
	    }
						
	    status = readSSRAM(chkbuf + j * bankSize, j * bankSize, bankSize, dma);
	    if (status != ADMXRC2_SUCCESS) {
	      printf("Error: failed to read SSRAM\n");
	      failed = TRUE;
	      break;
	    }
	  }
					
	  for (j = 0; j < numBank; j++) {
	    if (!(bankTestMask & (1U << j))) {
	      continue;
	    }
						
	    for (k = 0; k < bankSize; k++) {
	      unsigned long idx = j * bankSize + k;
							
	      act = chkbuf[idx];
	      if (exp != act) {
		if (nError == 10) {
		  printf("[more errors...]\n");
		} else if (nError < 10) {
		  printf("Error at index 0x%8.8lx: expected/actual = 0x%8.8lx/0x%8.8lx\n",
			 (unsigned long) idx, (unsigned long) exp, (unsigned long) act);
		}
		nError++;
	      }
	    }
	  }
	}
      }
			
      /* Perform own address test - look for shorted/stuck address pins */
      if (!failed && nError == 0) {
	printf("Writing own address...\n");
				
	for (j = 0; j < memorySize; j++) {
	  rambuf[j] = j;
	}
				
	for (j = 0; j < numBank; j++) {
	  if (!(bankTestMask & (1U << j))) {
	    continue;
	  }
					
	  status = writeSSRAM(rambuf + j * bankSize, j * bankSize, bankSize, dma);
	  if (status != ADMXRC2_SUCCESS) {
	    printf("Error: failed to write SSRAM\n");
	    failed = TRUE;
	    break;
	  }
	}
				
	if (!failed) {
	  for (j = 0; j < numBank; j++) {
	    if (!(bankTestMask & (1U << j))) {
	      continue;
	    }
						
	    status = readSSRAM(chkbuf + j * bankSize, j * bankSize, bankSize, dma);
	    if (status != ADMXRC2_SUCCESS) {
	      printf("Error: failed to read SSRAM\n");
	      failed = TRUE;
	      break;
	    }
	  }
					
	  for (j = 0; j < numBank; j++) {
	    if (!(bankTestMask & (1U << j))) {
	      continue;
	    }
						
	    for (k = 0; k < bankSize; k++) {
	      unsigned long idx = j * bankSize + k;
							
	      exp = idx;
	      act = chkbuf[idx];
	      if (exp != act) {
		if (nError == 100) {
		  printf("[more errors...]\n");
		} else if (nError < 100) {
		  printf("Error at index 0x%8.8lx: expected/actual = 0x%8.8lx/0x%8.8lx\n",
			 (unsigned long) idx, (unsigned long) exp, (unsigned long) act);
		}
		nError++;
	      }
	    }
	  }
	}
      }
			
      /* Perform own bit-reversed address test - look for shorted/stuck address pins */
      if (!failed && nError == 0) {
	printf("Writing own address bit-reversed...\n");
				
	for (j = 0; j < memorySize; j++) {
	  rambuf[j] = reverse(j);
	}
				
	for (j = 0; j < numBank; j++) {
	  if (!(bankTestMask & (1U << j))) {
	    continue;
	  }
					
	  status = writeSSRAM(rambuf + j * bankSize, j * bankSize, bankSize, dma);
	  if (status != ADMXRC2_SUCCESS) {
	    printf("Error: failed to write SSRAM\n");
	    failed = TRUE;
	    break;
	  }
	}
				
	if (!failed) {
	  for (j = 0; j < numBank; j++) {
	    if (!(bankTestMask & (1U << j))) {
	      continue;
	    }
						
	    status = readSSRAM(chkbuf + j * bankSize, j * bankSize, bankSize, dma);
	    if (status != ADMXRC2_SUCCESS) {
	      printf("Error: failed to read SSRAM\n");
	      failed = TRUE;
	      break;
	    }
	  }
					
	  for (j = 0; j < numBank; j++) {
	    if (!(bankTestMask & (1U << j))) {
	      continue;
	    }
						
	    for (k = 0; k < bankSize; k++) {
	      unsigned long idx = j * bankSize + k;
							
	      exp = reverse(idx);
	      act = chkbuf[idx];
	      if (exp != act) {
		if (nError == 100) {
		  printf("[more errors...]\n");
		} else if (nError < 100) {
		  printf("Error at index 0x%8.8lx: expected/actual = 0x%8.8lx/0x%8.8lx\n",
			 (unsigned long) idx, (unsigned long) exp, (unsigned long) act);
		}
		nError++;
	      }
	    }
	  }
	}
      }
			
      if (!failed && nError == 0) {
	printf("PASSED\n");
      } else {
	printf("*** FAILED with %ld error(s)\n", (long) nError);
	break;
      }
    }
		
    if (failed) {
      break;
    }
  }
	
  if (speed) {
    /*
    ** Measure the host read and host write speeds of the SSRAM,
    ** using DMA.
    */
    float          bytes, rate;
    double         delta;
    struct timeval start, now, elapsed;
		
    printf("Measuring access speed...\n");
		
    start = get_time();
    bytes = 0.0f;
    do {
      readSSRAM(rambuf, 0, memorySize, dma);
      bytes += (float) (memorySize * sizeof(uint32_t));
      now = get_time();
      elapsed = time_subtract(&now, &start);
      delta = time_to_double(&elapsed);
    } while (delta < 2.0);
		
    rate = bytes / delta;
    printf("SSRAM to host throughput = %.1fMB/s\n", rate / 1048576.0f);
		
    start = get_time();
    bytes = 0.0f;
    do {
      writeSSRAM(rambuf, 0, memorySize, dma);
      bytes += (float) (memorySize * sizeof(uint32_t));
      now = get_time();
      elapsed = time_subtract(&now, &start);
      delta = time_to_double(&elapsed);
    } while (delta < 2.0);
		
    rate = bytes / delta;
    printf("Host to SSRAM throughput = %.1fMB/s\n", rate / 1048576.0f);
  }
	
 done:
  if (rambuf != NULL) {
    ADMXRC2_Free(rambuf);
  }

  if (chkbuf != NULL) {
    ADMXRC2_Free(chkbuf);
  }

  if (dmaDesc1Valid) {
    ADMXRC2_UnsetupDMA(card, dmaDesc1);
  }

  if (dmaDesc2Valid) {
    ADMXRC2_UnsetupDMA(card, dmaDesc2);
  }

  if (card != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(card);
  }
	
  return ret;
}
