--
-- plxddsm.vhd - PLX local bus demand mode DMA state machine
--
-- Modules defined:
--
--    plxddsm               PLX local bus demand mode DMA state machine
--
-- The plxddsm module defines the state machine and associated logic that
-- controls a single PLX DMA channel in demand mode.
--
-- One cycle of address decoding is permitted.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

entity plxddsm is
    port(
        clk:           in       std_logic;  -- Local bus clock
        rst:           in       std_logic;  -- Should reduce during synthesis to global reset
        sr:            in       std_logic;  -- Synchronous reset (do not assert unless bus is idle)
        qlads:         in       std_logic;  -- LADS# from local bus, qualified by address decoding
        ldack:         in       std_logic;  -- LDACK# signal from local bus, active high
        lblast:        in       std_logic;  -- LBLAST# signal from local bus, active high
        lwrite:        in       std_logic;  -- LWRITE signal from local bus
        ldreq:         out      std_logic;  -- To local bus LDREQ# signal, active high
        lready:        out      std_logic;  -- To local bus LREADY# signal, active high
        lready_oe:     out      std_logic;  -- Enable for LREADY# on local bus
        lbterm:        out      std_logic;  -- To local bus LBTERM# signal, active high
        lbterm_oe:     out      std_logic;  -- Enable for LBTERM# on local bus
        ld_oe:         out      std_logic;  -- Output enable for LD pins
        request:       in       std_logic;  -- Pulse to request cycle
        burst:         in       std_logic;  -- Burst cycle requested / end burst
        transfer:      out      std_logic;  -- Data transfer on this clock edge, when asserted
        decode:        out      std_logic;  -- Application should do further address decoding, when asserted
        write:         out      std_logic); -- Cycle is a write, when asserted
end;

architecture rtl of plxddsm is
    
    type state_type is (s_idle, s_req, s_decode, s_xfer);
    signal state, n_state : state_type;
    signal i_transfer, n_transfer : std_logic;
    signal i_lready, n_lready : std_logic;
    signal i_lready_oe, n_lready_oe : std_logic;
    signal i_lbterm, n_lbterm : std_logic;
    signal i_ld_oe, n_ld_oe : std_logic;
    signal i_decode, n_decode : std_logic;
    signal burst_q, n_burst_q : std_logic;
    signal i_write, n_write : std_logic;
    
begin
    
    lready     <= i_lready;
    lready_oe  <= i_lready_oe;
    lbterm     <= i_lbterm;
    lbterm_oe  <= i_lready_oe;
    ld_oe      <= i_ld_oe;
    transfer   <= i_transfer;
    decode     <= i_decode;
    write      <= i_write;

    --
    -- Generate the demand-mode DMA request signal.
    --
    generate_ldreq: process(
        state,
        ldack,
        burst,
        burst_q)
    begin
        if  (state = s_req and ldack = '0') or
            (state = s_req and ldack = '1' and (burst_q and burst) = '1') or
            ((state = s_xfer or state = s_decode) and (burst_q and burst) = '1') then
            ldreq <= '1';
        else
            ldreq <= '0';
        end if;
    end process;
    
    --
    -- Generate the current state.
    --
    current_state: process(clk, rst)
    begin
        if rst = '1' then
            state        <= s_idle;
            i_transfer   <= '0';
            i_lready     <= '0';
            i_lready_oe  <= '0';
            i_lbterm     <= '0';
            i_ld_oe      <= '0';
            i_write      <= '0';
            i_decode     <= '0';
            burst_q      <= '0';
        elsif clk'event and clk = '1' then
            if sr = '1' then
                state        <= s_idle;
                i_transfer   <= '0';
                i_lready     <= '0';
                i_lready_oe  <= '0';
                i_lbterm     <= '0';
                i_ld_oe      <= '0';
                i_write      <= '0';
                i_decode     <= '0';
                burst_q      <= '0';
            else
                state        <= n_state;
                i_transfer   <= n_transfer;
                i_lready     <= n_lready;
                i_lready_oe  <= n_lready_oe;
                i_lbterm     <= n_lbterm;
                i_ld_oe      <= n_ld_oe;
                i_write      <= n_write;
                i_decode     <= n_decode;
                burst_q      <= n_burst_q;
            end if;
        end if;
    end process;
    
    --
    -- Determine next state.
    --
    next_state: process(
        request,
        burst,
        burst_q,
        state,
        qlads,
        lblast,
        lwrite,
        ldack,
        i_write)
    begin
        case state is 
            when s_idle =>
                n_transfer   <= '0';
                n_burst_q    <= burst;
                n_decode     <= '0';
                n_lready     <= '0';
                n_lbterm     <= '0';
                n_ld_oe      <= '0';
                n_lready_oe  <= '0';
                if request = '1' then
                    n_state      <= s_req;
                else
                    n_state      <= s_idle;
                end if;
                n_write      <= i_write;

            when s_req =>
                n_transfer   <= '0';
                n_burst_q    <= burst_q and burst;
                n_lbterm     <= '0';
                n_ld_oe      <= '0';
                n_lready     <= '0';
                n_lready_oe  <= '0';
                if qlads = '1' and ldack = '1' then
                    n_state      <= s_decode;
                    n_decode     <= '1';
                    n_write      <= lwrite;
                else
                    n_state      <= s_req;
                    n_decode     <= '0';
                    n_write      <= i_write;
                end if;
                    
            when s_decode =>
                n_state      <= s_xfer;
                n_burst_q    <= burst_q and burst;
                n_decode     <= '0';
                n_lbterm     <= '0';
                n_ld_oe      <= not i_write;
                n_lready     <= '1';
                n_lready_oe  <= '1';
                n_transfer   <= '1';
                n_write      <= i_write;
                
            when s_xfer =>
                n_burst_q    <= burst_q and burst;
                n_decode     <= '0';
                n_lbterm     <= '0';
                n_lready_oe  <= '1';
                n_write      <= i_write;
                if lblast = '1' then
                    n_state      <= s_idle;
                    n_ld_oe      <= '0';
                    n_lready     <= '0';
                    n_transfer   <= '0';
                else
                    n_state      <= s_xfer;
                    n_ld_oe      <= not i_write;
                    n_lready     <= '1';
                    n_transfer   <= '1';
                end if;
        end case;
    end process;
    
end rtl;
