--
-- plxdmsm.vhd - PLX local bus direct master state machine
--
-- Modules defined:
--
--    plxdmsm              PLX local bus direct master state machine
--
-- This state machine is designed to initiate direct master transfers on
-- the PLX local bus.
--

library ieee;
use ieee.std_logic_1164.all;

entity plxdmsm is
    port(
        clk:           in       std_logic;  -- Local bus clock
        rst:           in       std_logic;  -- Asynchronous reset, should reduce to global reset
        sr:            in       std_logic;  -- Synchronous reset, must be asserted when bus idle
        request:       in       std_logic;  -- Pulse to request a direct-master cycle
        write:         in       std_logic;  -- Assert coincident with 'request' for a write cycle
        stop:          in       std_logic;  -- Pulse to terminate burst
        transfer:      out      std_logic;  -- Asserted 1 clock after data transfer takes place
        done:          out      std_logic;  -- Asserted when machine is idle
        lready:        in       std_logic;  -- Local bus LREADY# signal (PCI9656) or LREADYO# (PCI9080) signal
        lbterm:        in       std_logic;  -- Local bus LREADY# signal (PCI9656) or LREADYO# (PCI9080) signal
        fholda:        in       std_logic;  -- From local bus arbiter, FPGA's bus grant signal
        fhold:         out      std_logic;  -- To local bus arbiter, FPGA's bus request signal
        la_oe:         out      std_logic;  -- Enable for local bus LA signal
        lads:          out      std_logic;  -- To local bus LADS# signal, active high
        lads_oe:       out      std_logic;  -- Enable for local bus LADS# signal
        lbe_oe:        out      std_logic;  -- Enable for local bus LBE# signal
        lblast:        out      std_logic;  -- To local bus LBLAST# signal, active high
        lblast_oe:     out      std_logic;  -- Enable for local bus LBLAST# signal
        ld_oe:         out      std_logic;  -- Enable for local bus LD signal
        lwrite:        out      std_logic;  -- To local bus LWRITE signal
        lwrite_oe:     out      std_logic); -- Enable for local bus LWRITE signal
end plxdmsm;

architecture rtl of plxdmsm is

    type state_type is (s_idle, s_arb, s_addr, s_xfer);
    signal state, n_state : state_type;
    signal i_la_oe, n_la_oe : std_logic;
    signal i_lads, n_lads : std_logic;
    signal i_lads_oe, n_lads_oe : std_logic;
    signal i_lbe_oe, n_lbe_oe : std_logic;
    signal i_lblast, n_lblast : std_logic;
    signal i_lblast_oe, n_lblast_oe : std_logic;
    signal i_ld_oe, n_ld_oe : std_logic;
    signal i_lwrite, n_lwrite : std_logic;
    signal i_lwrite_oe, n_lwrite_oe : std_logic;
    signal i_done, n_done : std_logic;
    signal i_fhold, n_fhold : std_logic;
    signal i_transfer, n_transfer : std_logic;
    
    signal stopping, n_stopping : std_logic;
    signal write_q, n_write_q : std_logic;
    
begin
    
    --
    -- Generate outputs.
    --
    la_oe     <= i_la_oe;
    lads      <= i_lads;
    lads_oe   <= i_lads_oe;
    lbe_oe    <= i_lbe_oe;
    lblast    <= i_lblast;
    lblast_oe <= i_lblast_oe;
    ld_oe     <= i_ld_oe;
    lwrite    <= i_lwrite;
    lwrite_oe <= i_lwrite_oe;
    done      <= i_done;
    fhold     <= i_fhold;
    transfer  <= i_transfer;

    --
    -- Generate the current state.
    --
    state_transition : process(clk, rst)
    begin
        if rst = '1' then
            state <= s_idle;
            i_la_oe <= '0';
            i_lads <= '0';
            i_lads_oe <= '0';
            i_lbe_oe <= '0';
            i_lblast <= '0';
            i_lblast_oe <= '0';
            i_ld_oe <= '0';
            i_lwrite <= '0';
            i_lwrite_oe <= '0';
            i_fhold <= '0';
            i_done <= '1';
            i_transfer <= '0';
            stopping <= '0';
            write_q <= '0';
        elsif clk'event and clk = '1' then
            if sr = '1' then
                state <= s_idle;
                i_la_oe <= '0';
                i_lads <= '0';
                i_lads_oe <= '0';
                i_lbe_oe <= '0';
                i_lblast <= '0';
                i_lblast_oe <= '0';
                i_ld_oe <= '0';
                i_lwrite <= '0';
                i_lwrite_oe <= '0';
                i_fhold <= '0';
                i_done <= '1';
                i_transfer <= '0';
                stopping <= '0';
                write_q <= '0';
            else
                state <= n_state;
                i_la_oe <= n_la_oe;
                i_lads <= n_lads;
                i_lads_oe <= n_lads_oe;
                i_lbe_oe <= n_lbe_oe;
                i_lblast <= n_lblast;
                i_lblast_oe <= n_lblast_oe;
                i_ld_oe <= n_ld_oe;
                i_lwrite <= n_lwrite;
                i_lwrite_oe <= n_lwrite_oe;
                i_fhold <= n_fhold;
                i_done <= n_done;
                i_transfer <= n_transfer;
                stopping <= n_stopping;
                write_q <= n_write_q;
            end if;
        end if;
    end process;
    
    --
    -- Determine next state.
    --
    next_state : process(
        state, 
        request,
        stop,
        write,
        write_q,
        stopping,
        fholda,
        lready,
        lbterm,
        i_lblast)
    begin
        case state is
            when s_idle =>
                n_ld_oe <= '0';
                n_transfer <= '0';
                if request = '1' then
                    n_state <= s_arb;
                    n_fhold <= '1';
                    n_lads <= '0';
                    n_lads_oe <= '0';
                    n_la_oe <= '0';
                    n_lbe_oe <= '0';
                    n_lblast <= '0';
                    n_lblast_oe <= '0';
                    n_lwrite <= '0';
                    n_lwrite_oe <= '0';
                    n_done <= '0';
                    n_stopping <= stop;
                    n_write_q <= write;                  
                else
                    n_state <= s_idle;
                    n_fhold <= '0';
                    n_lads <= '0';
                    n_lads_oe <= '0';
                    n_lblast <= '0';
                    n_lblast_oe <= '0';
                    n_la_oe <= '0';
                    n_lbe_oe <= '0';
                    n_lwrite <= '0';
                    n_lwrite_oe <= '0';
                    n_done <= '1';
                    n_stopping <= '0';
                    n_write_q <= write_q;                  
                end if;
                
            when s_arb =>
                n_done <= '0';
                n_fhold <= '1';
                n_ld_oe <= '0';
                n_stopping <= stop or stopping;
		n_transfer <= '0';
                n_write_q <= write_q;
                if fholda = '1' then
                    n_state <= s_addr;
                    n_lads <= '1';
                    n_lads_oe <= '1';
                    n_la_oe <= '1';
                    n_lbe_oe <= '1';
                    n_lblast <= '0';
                    n_lblast_oe <= '1';
                    n_lwrite <= write_q;
                    n_lwrite_oe <= '1';
                else
                    n_state <= s_arb;
                    n_lads <= '0';
                    n_lads_oe <= '0';
                    n_la_oe <= '0';
                    n_lbe_oe <= '0';
                    n_lblast <= '0';
                    n_lblast_oe <= '0';
                    n_lwrite <= '0';
                    n_lwrite_oe <= '0';
                end if;
                
            when s_addr =>
                n_state <= s_xfer;
                n_done <= '0';
                n_fhold <= '1';
                n_lads <= '0';
                n_lads_oe <= '1';
                n_la_oe <= '1';
                n_lbe_oe <= '1';
                n_lblast <= stop or stopping;
                n_lblast_oe <= '1';
                n_ld_oe <= write_q;
                n_lwrite <= write_q;
                n_lwrite_oe <= '1';
                n_stopping <= stop or stopping;
		n_transfer <= '0';
                n_write_q <= write_q;
                
            when s_xfer =>
                n_fhold <= '1';
                n_stopping <= stop or stopping;
                n_write_q <= write_q;
                if (i_lblast = '1' and lready = '1') or lbterm = '1' then
                    n_state <= s_idle;
                    n_done <= '1';
                    n_lads <= '0';
                    n_lads_oe <= '0';
                    n_la_oe <= '0';
                    n_lbe_oe <= '0';
                    n_lblast <= '0';
                    n_lblast_oe <= '1';
                    n_ld_oe <= '0';
                    n_lwrite <= '0';
                    n_lwrite_oe <= '1';
                    n_transfer <= '1';
                else
                    n_state <= s_xfer;
                    n_done <= '0';
                    n_lads <= '0';
                    n_lads_oe <= '1';
                    n_la_oe <= '1';
                    n_lbe_oe <= '1';
                    n_lblast <= stop or stopping;
                    n_lblast_oe <= '1';
                    n_ld_oe <= write_q;
                    n_lwrite <= write_q;
                    n_lwrite_oe <= '1';
                    n_transfer <= lready;
                end if;
        end case;
    end process;
    
end rtl;

