--
-- ddma-xrc2pl.vhd - example demand-mode DMA design
--
-- (C) Copyright Alpha Data Parallel Systems Ltd. 2001-2004
--
-- Example program for:
--
--   o ADM-XRCIIPro-Lite
--   o ADM-XRCIIPro
--
-- Demand mode DMA is used to read data from the host via DMA channel 0,
-- into a FIFO. The data read from the FIFO and written back to the host via
-- DMA channel 1, also in demand-mode.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.sdkcomp.all;

entity ddma is
    port(
        lclk:            in        std_logic;
        lreset_l:        in        std_logic;
        lwrite:          in        std_logic;
        lads_l:          in        std_logic;
        lblast_l:        in        std_logic;
        lbterm_l:        out       std_logic;
        lad:             inout     std_logic_vector(31 downto 0);
        lready_l:        out       std_logic;
        lbe_l:           in        std_logic_vector(3 downto 0);
        fholda:          in        std_logic;
        ldreq_l:         out       std_logic_vector(1 downto 0);
        ldack_l:         in        std_logic_vector(1 downto 0));
end ddma;

architecture mixed of ddma is

    signal rst:         std_logic;
    signal lads_i:      std_logic;
    signal lblast_i:    std_logic;
    signal lbe_i:       std_logic_vector(3 downto 0);
    signal lwrite_i:    std_logic;
    signal ldack_i:     std_logic_vector(1 downto 0);
    signal ld_i:        std_logic_vector(31 downto 0);
    signal la_i:        std_logic_vector(23 downto 2);
    signal lready_o:    std_logic;
    signal lready_oe:   std_logic;
    signal lbterm_o:    std_logic;
    signal lbterm_oe:   std_logic;
    signal ldreq_o:     std_logic_vector(1 downto 0);
    signal ld_o:        std_logic_vector(31 downto 0);
    signal ld_oe:       std_logic;
    
    signal la_q:        std_logic_vector(23 downto 2);

    signal dd_qlads:    std_logic;
    signal ds_qlads:    std_logic;
    signal dd0_write:   std_logic;
    signal dd1_write:   std_logic;
    signal ds_write:    std_logic;
    signal dd0_xfer:    std_logic;
    signal dd1_xfer:    std_logic;
    signal ds_xfer:     std_logic;
    signal dd0_lrdy:    std_logic;
    signal dd1_lrdy:    std_logic;
    signal ds_lrdy:     std_logic;
    signal dd0_lrdy_oe: std_logic;
    signal dd1_lrdy_oe: std_logic;
    signal ds_lrdy_oe:  std_logic;
    signal dd0_lbt:     std_logic;
    signal dd1_lbt:     std_logic;
    signal ds_lbt:      std_logic;
    signal dd0_lbt_oe:  std_logic;
    signal dd1_lbt_oe:  std_logic;
    signal ds_lbt_oe:   std_logic;

    signal fifo_e:      std_logic;
    signal fifo_ne:     std_logic;
    signal fifo_read:   std_logic;
    signal fifo_f:      std_logic;
    signal fifo_nf:     std_logic;
    signal fifo_write:  std_logic;
    signal fifo_status: std_logic_vector(34 downto 0);

    signal dd0_req:     std_logic;
    signal dd1_req:     std_logic;
    signal dd0_burst:   std_logic;
    signal dd1_burst:   std_logic;

    signal icount:      std_logic_vector(29 downto 0);
    signal icount0:     std_logic;
    signal icount4:     std_logic;
    signal wrote_icount: std_logic;
    signal ocount:      std_logic_vector(29 downto 0);
    signal ocount0:     std_logic;
    signal ocount4:     std_logic;
    signal wrote_ocount: std_logic;

    signal logic0:      std_logic;
    signal logic1:      std_logic;

begin
    
    logic0 <= '0';
    logic1 <= '1';

    --
    -- Convert the inputs to active high.
    --	
    rst        <= not lreset_l;
    lads_i     <= not lads_l;
    lblast_i   <= not lblast_l;
    lbe_i      <= not lbe_l;
    lwrite_i   <= lwrite;
    ld_i       <= lad;
    la_i       <= lad(23 downto 2);
    ldack_i    <= not ldack_l;

    --
    -- Output active low signals.
    --
    ldreq_l    <= not ldreq_o;

    --
    -- Generate qualified versions of 'lads_l'.
    --
    -- For direct slave access, this is when 'lads_l' is asserted AND the FPGA
    -- is being addressed AND the FPGA is not the local bus master AND the
    -- cycle is not demand-mode DMA.
    --
    -- For demand-mode DMA access, this is when 'lads_l' is asserted AND the
    -- FPGA is being addressed and the FPGA is not the local bus master. The
    -- demand-mode DMA acknowledge signals are checked within the 'plxddsm'
    -- module.
    --
    ds_qlads <= lads_i and not la_i(23) and not fholda and not OR_reduce(ldack_i);
    dd_qlads <= lads_i and not la_i(23) and not fholda;

    --
    -- 'lbterm_l' should only be driven when the FPGA is addressed; otherwise
    -- float, because the control logic on the card might also drive it.
    --
    -- The individual signals from 3 sources are ORed together here:
    --
    --   o bterm from DMA channel 0 demand-mode DMA state machine
    --   o bterm from DMA channel 1 demand-mode DMA state machine
    --   o bterm from direct slave state machine
    --
    lbterm_o   <= dd0_lbt or dd1_lbt or ds_lbt;
    lbterm_oe  <= dd0_lbt_oe or dd1_lbt_oe or ds_lbt_oe;
    lbterm_l   <= not lbterm_o when lbterm_oe = '1' else 'Z';

    --
    -- 'lready_l' should only be driven when the FPGA is addressed; otherwise
    -- float because the control logic on the card might also drive it.
    --
    -- The individual signals from 3 sources are ORed together here:
    --
    --   o ready from DMA channel 0 demand-mode DMA state machine
    --   o ready from DMA channel 1 demand-mode DMA state machine
    --   o ready from direct slave state machine
    --
    lready_o  <= dd0_lrdy or dd1_lrdy or ds_lrdy;
    lready_oe <= dd0_lrdy_oe or dd1_lrdy_oe or ds_lrdy_oe;
    lready_l  <= not lready_o when lready_oe = '1' else 'Z';

    --
    -- Drive the local data bus on a read.
    --
    lad <= ld_o when ld_oe = '1' else (others => 'Z');

    --
    -- Latch 'lad' on assertion of 'lads_l', in order to improve timing. There
    -- is no need to increment the address, because we don't allow bursting for
    -- register addresses, and DMA transfers that burst are expected to use
    -- constant local addressing.
    --
    latch_address : process(rst, lclk)
    begin
        if rst = '1' then
            la_q <= (others => '0');
        elsif lclk'event and lclk = '1' then
            if lads_i = '1' then
                la_q <= la_i;
            end if;
        end if;
    end process;

    --
    -- This process implements the DMA transfer count registers and flags for
    -- each direction.
    --
    --    icount          number of words remaining in PCI-to-local DMA transfer
    --    icount0         '1' => 'icount' is zero
    --    icount4         '1' => 'icount' is less than or equal to 4
    --    ocount          number of words remaining in local-to-PCI DMA transfer
    --    ocount0         '1' => 'ocount' is zero
    --    ocount4         '1' => 'ocount' is less than or equal to 4
    --
    -- Prior to the host kicking off a demand-mode DMA transfer, the host must
    -- program the transfer count registers must be programmed with the length
    -- of the DMA transfer being performed.
    --
    gen_counts: process(rst, lclk)
    begin
        if rst = '1' then
            icount  <= (others => '0');
            icount4 <= '1';
            icount0 <= '1';
            wrote_icount <= '0';
            ocount  <= (others => '0');
            ocount4 <= '1';
            ocount0 <= '1';
            wrote_ocount <= '0';
        elsif lclk'event and lclk = '1' then
            if ds_xfer = '1' and ds_write = '1' and la_q(2) = '0' then
                if lbe_i(0) = '1' then
                    icount(5 downto 0) <= ld_i(7 downto 2);
                end if;
                if lbe_i(1) = '1' then 
                    icount(13 downto 6) <= ld_i(15 downto 8);
                end if;
                if lbe_i(2) = '1' then 
                    icount(21 downto 14) <= ld_i(23 downto 16);
                end if;
                if lbe_i(3) = '1' then 
                    icount(29 downto 22) <= ld_i(31 downto 24);
                end if;
            else
                if fifo_write = '1' then
                    icount <= icount - 1;
                end if;
            end if;

            if ds_xfer = '1' and ds_write = '1' and la_q(2) = '1' then
                if lbe_i(0) = '1' then 
                    ocount(5 downto 0) <= ld_i(7 downto 2);
                end if;
                if lbe_i(1) = '1' then 
                    ocount(13 downto 6) <= ld_i(15 downto 8);
                end if;
                if lbe_i(2) = '1' then 
                    ocount(21 downto 14) <= ld_i(23 downto 16);
                end if;
                if lbe_i(3) = '1' then 
                    ocount(29 downto 22) <= ld_i(31 downto 24);
                end if;
            else
                if fifo_read = '1' then
                    ocount <= ocount - 1;
                end if;
            end if;

            wrote_icount <= ds_xfer and ds_write and not la_q(2);
            if wrote_icount = '1' then
                if icount(29 downto 4) = 0 and icount(3 downto 0) <= 4 then
                    icount4 <= '1';
                else
                    icount4 <= '0';
                end if;
                if icount = 0 then
                    icount0 <= '1';
                else
                    icount0 <= '0';
                end if;
            else
                if fifo_write = '1' then
                    if icount = 5 then
                        icount4 <= '1';
                    end if;
                    if icount = 1 then
                        icount0 <= '1';
                    end if;
                end if;
            end if;

            wrote_ocount <= ds_xfer and ds_write and la_q(2);
            if wrote_ocount = '1' then
                if ocount(29 downto 4) = 0 and ocount(3 downto 0) <= 4 then
                    ocount4 <= '1';
                else
                    ocount4 <= '0';
                end if;
                if ocount = 0 then
                    ocount0 <= '1';
                else
                    ocount0 <= '0';
                end if;
            else
                if fifo_read = '1' then
                    if ocount = 5 then
                        ocount4 <= '1';
                    end if;
                    if ocount = 1 then
                        ocount0 <= '1';
                    end if;
                end if;
            end if;
        end if;
    end process;

    --
    -- For PCI->local direction, we can wait until the FIFO is not nearly-full
    -- before requesting a demand-mode DMA burst. We need to pause if the FIFO
    -- becomes nearly-full.
    --
    dd0_req   <= not fifo_nf and not icount0;
    dd0_burst <= not fifo_nf;

    --
    -- For the local->PCI direction, we wait until the FIFO is not empty
    -- before requesting a demand-mode DMA burst. We need to pause if the FIFO
    -- becomes empty.
    --
    -- NOTE: the bug that afflicts demand-mode DMA on the PCI9656 does not
    --       exist in the ADM-XPL; hence the simpler code here (although the
    --       code used for the ADM-XRCII version would also work with the
    --       ADM-XPL).
    --
    dd1_req    <= not fifo_e and not ocount0;
    dd1_burst  <= not fifo_ne;

    --
    -- Writing into the FIFO only happens when data is written on
    -- PLX DMA channel 0.
    --
    fifo_write <= dd0_xfer and dd0_write;

    --
    -- Reading from the FIFO only happens when data is read on PLX
    -- DMA channel 1.
    --
    fifo_read  <= dd1_xfer and not dd1_write;

    --
    -- Instantiate the demand-mode DMA state machine for DMA channel 0.
    -- Monitors the local bus for channel 0 demand-mode DMA cycles and
    -- responds appropriately.
    --
    plxddsm0: plxddsm
        port map(
            clk        => lclk,
            rst        => rst,
            sr         => logic0,
            qlads      => dd_qlads,
            ldack      => ldack_i(0),
            lblast     => lblast_i,
            lwrite     => lwrite_i,
            ld_oe      => open,
            lready     => dd0_lrdy,
            lready_oe  => dd0_lrdy_oe,
            lbterm     => dd0_lbt,
            lbterm_oe  => dd0_lbt_oe,
            ldreq      => ldreq_o(0),
            transfer   => dd0_xfer,
            decode     => open,
            request    => dd0_req,
            burst      => dd0_burst,
            write      => dd0_write);

    --
    -- Instantiate the demand-mode DMA state machine for DMA channel 1.
    -- Monitors the local bus for channel 1 demand-mode DMA cycles and
    -- responds appropriately.
    --
    plxddsm1: plxddsm
        port map(
            clk        => lclk,
            rst        => rst,
            sr         => logic0,
            qlads      => dd_qlads,
            ldack      => ldack_i(1),
            lblast     => lblast_i,
            lwrite     => lwrite_i,
            ld_oe      => ld_oe,
            lready     => dd1_lrdy,
            lready_oe  => dd1_lrdy_oe,
            lbterm     => dd1_lbt,
            lbterm_oe  => dd1_lbt_oe,
            ldreq      => ldreq_o(1),
            transfer   => dd1_xfer,
            decode     => open,
            request    => dd1_req,
            burst      => dd1_burst,
            write      => dd1_write);

    --
    -- Instantiate the direct slave state machine; monitors the local bus for
    -- direct slave cycles and responds appropriately.
    --
    plxdssm0: plxdssm
        port map(
            clk        => lclk,
            rst        => rst,
            sr         => logic0,
            qlads      => ds_qlads,
            lblast     => lblast_i,
            lwrite     => lwrite_i,
            ld_oe      => open,
            lready     => ds_lrdy,
            lready_oe  => ds_lrdy_oe,
            lbterm     => ds_lbt,
            lbterm_oe  => ds_lbt_oe,
            transfer   => ds_xfer,
            decode     => open,
            write      => ds_write,
            ready      => logic1,
            stop       => logic1);

    --
    -- The value of 4 for the parameters 'n_full' and 'n_empty' means
    -- that nearly-full => 4 or less free locations in the FIFO, and
    -- nearly-empty => 4 or less words present the FIFO.
    --    
    sfifo0: sfifo
	generic map(
            n_full     => 4,
            n_empty    => 4)
        port map(
            status     => fifo_status,
            clk        => lclk,
            rst        => rst,
            sr         => logic0,
            d          => ld_i,
            q          => ld_o,
            w          => fifo_write,
            we         => open,
            wnf        => fifo_nf,
            wf         => fifo_f,
            r          => fifo_read,
            re         => fifo_e,
            rf         => open,
            rne        => fifo_ne);

end mixed;
